#!/usr/bin/env python3
"""
Create a snapshot from the Compadre repository containing this script into the
Trilinos repository pointed to, omitting the kokkos, kokkos-kernels, pycompadre,
and scripts directories.

Note:  This will assert that both repositories are in a clean state.

Warning:  This will clean out any locally-ignored files in Compadre (e.g.,
    ignored via .git/info/exclude) to avoid copying them over and then
    committing them to Trilinos.  Be sure you don't have any ignored files in
    Compadre that you want to keep before running this script.

To test this script to ensure that it's working correctly, simply use

    pytest snapshot_into_trilinos.py
"""
import os
import pytest
import sys




def parse_arguments(argv):
    """
    Parse the command line arguments to the script.

    Parameters:
        argv (list):  The command line arguments to be parsed.

    Returns:
        dict:  A mapping from the options to their values.
    """
    import argparse, textwrap
    width = 79
    description = __doc__
    description = ("[ Snapshot Compadre Into Trilinos ]".center(width, "-") +
                   "\n" + description)
    examples = """
        Show what's going to happen without actually doing the snapshot::

            ./snapshot_into_trilinos.py \\
                --trilinos-dir /path/to/Trilinos \\
                --dry-run

        Actually do the snapshot::

            ./snapshot_into_trilinos.py \\
                --trilinos-dir /path/to/Trilinos
    """
    examples = textwrap.dedent(examples)
    examples = "[ Examples ]".center(width, "-") + "\n\n" + examples
    parser = argparse.ArgumentParser(
        description=description,
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument("-t", "--trilinos-dir", dest="trilinos_dir",
                        action="store", required=True, default=None,
                        help="The path (relative or absolute) to the root of "
                        "the Trilinos repository you wish to snapshot "
                        "Compadre into.")
    parser.add_argument("-d", "--dry-run", dest="dry_run", action="store_true",
                        default=False, help="Show what will happen, but don't "
                        "actually do it.")
    return parser.parse_args(argv)



def test_parse_arguments():
    """
    Test that the :func:`parse_arguments` function works as intended.
    """
    options = parse_arguments("--trilinos-dir /path/to/Trilinos --dry-run".
                              split())
    assert (options.trilinos_dir == "/path/to/Trilinos" and
            options.dry_run == True)
    options = parse_arguments("-t /path/to/Trilinos -d".split())
    assert (options.trilinos_dir == "/path/to/Trilinos" and
            options.dry_run == True)
    options = parse_arguments("-t /some/other/dir".split())
    assert (options.trilinos_dir == "/some/other/dir" and
            options.dry_run == False)



def check_trilinos_dir_is_root(trilinos_dir):
    """
    Check to ensure the Trilinos directory specified is indeed the root of the
    Trilinos repository by ensuring it contains a ``ProjectName.cmake`` that
    specifies Trilinos as the project name.
    """
    import re
    try:
        with open(os.path.join(trilinos_dir, "ProjectName.cmake"), "r") as f:
            data = f.read().lower()
            if not re.search(r"set\s*\(\s*project_name\s+trilinos\s*\)", data):
                sys.exit(f"Error:  {trilinos_dir}/ProjectName.cmake does not "
                         "set the project name to Trilinos.  Make sure you "
                         "point to the root of the Trilinos repository.")
    except FileNotFoundError:
        sys.exit(f"Error:  {trilinos_dir} does not contain a "
                 "ProjectName.cmake.  Make sure you point to the root of the "
                 "Trilinos repository.")



def test_check_trilinos_dir_is_root(tmpdir):
    """
    Test that the :func:`check_trilinos_dir_is_root` function works as
    intended.
    """
    # Ensure the check fails if trilinos_dir doesn't contain ProjectName.cmake.
    trilinos_dir = tmpdir.mkdir("trilinos_dir")
    with pytest.raises(SystemExit):
        check_trilinos_dir_is_root(trilinos_dir)

    # Ensure the check fails if ProjectName.cmake doesn't set the project name
    # to Trilinos.
    file_to_check = trilinos_dir.join("ProjectName.cmake")
    file_to_check.write("Doesn't set PROJECT_NAME to Trilinos")
    with pytest.raises(SystemExit):
        check_trilinos_dir_is_root(trilinos_dir)

    # Ensure the check passes if ProjectName.cmake does set the project name to
    # Trilinos.
    file_to_check.write("SET(PROJECT_NAME Trilinos)")
    check_trilinos_dir_is_root(trilinos_dir)



def create_directory_variables(trilinos_dir, verbose=False):
    """
    Use the given path to Trilinos to create a handful of other variables
    pointing to:
    *  The python_utils directory, so we can import the SnapshotDir utility.
    *  The Compadre repository containing this script.
    *  The location to where Compadre will be shapshotted.

    Parameters:
        trilinos_dir (str):  The path to Trilinos given on the command line.
        verbose (bool):  Whether or not to print out details.

    Returns:
        tuple:  The original Compadre directory, and the location to be
            snapshotted to.
    """
    trilinos_dir = os.path.abspath(trilinos_dir)
    python_utils_dir = os.path.join(trilinos_dir, "cmake/tribits/python_utils")
    sys.path.append(python_utils_dir)
    script_dir = os.path.dirname(os.path.realpath(__file__))
    compadre_orig_dir = os.path.abspath(os.path.join(script_dir, ".."))
    compadre_dest_dir = os.path.join(trilinos_dir, "packages/compadre")
    if verbose:
        print(f"Trilinos repository root:  {trilinos_dir}\n"
              f"Using snapshot-dir.py from:  {python_utils_dir}\n"
              f"Snapshotting Compadre from:  {compadre_orig_dir}\n"
              f"                      into:  {compadre_dest_dir}")
    return (compadre_orig_dir, compadre_dest_dir)



def test_create_directory_variables(capfd):
    """
    Test that the :func:`create_directory_variables` function works as
    intended.
    """
    trilinos_dir = "/path/to/Trilinos"
    compadre_dir = os.path.abspath(os.path.join(os.path.dirname(
        os.path.realpath(__file__)), ".."))
    (orig_dir, dest_dir) = create_directory_variables(trilinos_dir)
    assert (orig_dir == compadre_dir and
            dest_dir == "/path/to/Trilinos/packages/compadre" and
            os.path.join(trilinos_dir, "cmake/tribits/python_utils") in
                sys.path)
    out, err = capfd.readouterr()
    assert out == ""
    create_directory_variables(trilinos_dir, verbose=True)
    out, err = capfd.readouterr()
    expected = (f"Trilinos repository root:  {trilinos_dir}\n"
                f"Using snapshot-dir.py from:  {trilinos_dir}/cmake/tribits/"
                    "python_utils\n"
                f"Snapshotting Compadre from:  {compadre_dir}\n"
                f"                      into:  {trilinos_dir}/packages/"
                    "compadre\n")
    assert out == expected



def create_snapshot_dir_args(orig_dir, dest_dir, dry_run=False):
    """
    Create the arguments to pass to the SnapshotDir utility from TriBITS.

    Parameters:
        orig_dir (str):  The path to the Compadre repository to be snapshotted.
        dest_dir (str):  The path to where Compadre will be snapshotted.
        dry_run (bool):  Whether or not to show what will happen without
            actually doing it.

    Returns:
        list:  The arguments to be passed to the utility.
    """
    args = (f"--orig-dir {orig_dir}/ "
            f"--dest-dir {dest_dir}/ "
            "--exclude kokkos kokkos-kernels pycompadre scripts "
            "setup.py "
            "MANIFEST.in "
            "cmake/detect_trilinos_opts.cmake "
            "examples/Python_3D_Convergence.py.in "
            "examples/__init__.py "
            "examples/test_pycompadre.py "
            "--clean-ignored-files-orig-dir")
    if dry_run:
        args += " --show-defaults"
    return args.split(" ")



def test_create_snapshot_dir_args():
    """
    Test that the :func:`create_snapshot_dir_args` function works as intended.
    """
    orig = "from_here"
    dest = "to_there"
    args = create_snapshot_dir_args(orig, dest)
    expected = (f"--orig-dir {orig}/ --dest-dir {dest}/ --exclude kokkos "
                "kokkos-kernels pycompadre scripts "
                "setup.py "
                "MANIFEST.in "
                "cmake/detect_trilinos_opts.cmake "
                "examples/Python_3D_Convergence.py.in "
                "examples/__init__.py "
                "examples/test_pycompadre.py "
                "--clean-ignored-files-orig-dir")
    assert args == expected.split()
    args = create_snapshot_dir_args(orig, dest, dry_run=True)
    expected += " --show-defaults"
    assert args == expected.split()



def snapshot(snapshot_dir_args):
    """
    Perform the snapshot using the SnapshotDir utility from TriBITS.

    Parameters:
        shapshot_dir_args (str):  The arguments to pass to SnapshotDir.

    Returns:
        bool:  Whether or not the snapshot was successful.

    Note:
        The SnapshotDir utility is already unit tested within TriBITS, so
        there's no need for an additional test here.
    """
    import SnapshotDir
    return SnapshotDir.snapshotDirMainDriver(snapshot_dir_args)



if __name__ == "__main__":
    options = parse_arguments(sys.argv[1:])
    check_trilinos_dir_is_root(options.trilinos_dir)
    (orig_dir, dest_dir) = create_directory_variables(options.trilinos_dir,
                                                      options.dry_run)
    success = snapshot(create_snapshot_dir_args(orig_dir, dest_dir,
                                                options.dry_run))
    return_code = 0 if success else 1
    sys.exit(return_code)



#! /usr/bin/env python

"""
Snapshot a project into another project and perform the necessary repo actions
to provide a commit message that can be used to trace back to the exact point
in the source repository.
"""

#todo:
#  Support svn
#  Allow renaming of the source dir in the destination path
#  Check if a new snapshot is necessary?
#

import sys

#check the version number so that there is a good error message when argparse is not available.
#This checks for exactly 2.7 which is bad, but it is a python 2 script and argparse was introduced
#in 2.7 which is also the last version of python 2. If this script is updated for python 3 this
#will need to change, but for now it is not safe to allow 3.x to run this.
if sys.version_info[:2] != (2, 7):
  print "Error snapshot requires python 2.7 detected version is %d.%d." % (sys.version_info[0], sys.version_info[1])
  sys.exit(1)

import subprocess, argparse, re, doctest, os, datetime, traceback

def parse_cmdline(description):
  parser = argparse.ArgumentParser(usage="snapshot.py [options] source destination", description=description)

  parser.add_argument("-n", "--no-commit", action="store_false", dest="create_commit", default=True,
                      help="Do not perform a commit or create a commit message.")
  parser.add_argument("-v", "--verbose", action="store_true", dest="verbose_mode", default=False,
                      help="Enable verbose mode.")
  parser.add_argument("-d", "--debug", action="store_true", dest="debug_mode", default=False,
                      help="Enable debugging output.")
  parser.add_argument("--no-validate-repo", action="store_true", dest="no_validate_repo", default=False,
                      help="Reduce the validation that the source and destination repos are clean to a warning.")
  parser.add_argument("--source-repo", choices=["git","none"], default="",
                      help="Type of repository of the source, use none to skip all repository operations.")
  parser.add_argument("--dest-repo", choices=["git","none"], default="",
                      help="Type of repository of the destination, use none to skip all repository operations.")

  parser.add_argument("source",      help="Source project to snapshot from.")
  parser.add_argument("destination", help="Destination to snapshot too.")

  options = parser.parse_args()
  options = validate_options(options)
  return options
#end parseCmdline

def validate_options(options):
  apparent_source_repo_type="none"
  apparent_dest_repo_type="none"

  #prevent user from accidentally giving us a path that rsync will treat differently than expected.
  options.source      = options.source.rstrip(os.sep)
  options.destination = options.destination.rstrip(os.sep)

  options.source      = os.path.abspath(options.source)
  options.destination = os.path.abspath(options.destination)
  
  if os.path.exists(options.source):
    apparent_source_repo_type, source_root = deterimine_repo_type(options.source)
  else:
    raise RuntimeError("Could not find source directory of %s." % options.source)
  options.source_root = source_root

  if not os.path.exists(options.destination):
    print "Could not find destination directory of %s so it will be created." % options.destination
    os.makedirs(options.destination)

  apparent_dest_repo_type, dest_root = deterimine_repo_type(options.destination)
  options.dest_root = dest_root

  #error on svn repo types for now
  if apparent_source_repo_type == "svn" or apparent_dest_repo_type == "svn":
    raise RuntimeError("SVN repositories are not supported at this time.")

  if options.source_repo == "":
    #source repo type is not specified to just using the apparent type.
    options.source_repo = apparent_source_repo_type
  else:
    if options.source_repo != "none" and options.source_repo != apparent_source_repo_type:
      raise RuntimeError("Specified source repository type of %s conflicts with determined type of %s" % \
        (options.source_repo, apparent_source_repo_type))

  if options.dest_repo == "":
    #destination repo type is not specified to just using the apparent type.
    options.dest_repo = apparent_dest_repo_type
  else:
    if options.dest_repo != "none" and options.dest_repo != apparent_dest_repo_type:
      raise RuntimeError("Specified destination repository type of %s conflicts with determined type of %s" % \
        (options.dest_repo, apparent_dest_repo_type))

  return options
#end validate_options

def run_cmd(cmd, options, working_dir="."):
  cmd_str = " ".join(cmd)
  if options.verbose_mode:
    print "Running command '%s' in dir %s." % (cmd_str, working_dir)

  proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=working_dir)
  proc_stdout, proc_stderr = proc.communicate()
  ret_val = proc.wait()

  if options.debug_mode:
    print "==== %s stdout start ====" % cmd_str
    print proc_stdout
    print "==== %s stdout end ====" % cmd_str
    print "==== %s stderr ====" % cmd_str
    print proc_stderr
    print "==== %s stderr ====" % cmd_str
 
  if ret_val != 0:
    raise RuntimeError("Command '%s' failed with error code %d. Error message:%s%s%sstdout:%s" % \
      (cmd_str, ret_val, os.linesep, proc_stderr, os.linesep, proc_stdout))

  return proc_stdout, proc_stderr
#end run_cmd

def deterimine_repo_type(location):
  apparent_repo_type = "none"

  while location != "":
    if os.path.exists(os.path.join(location, ".git")):
      apparent_repo_type = "git"
      break
    elif os.path.exists(os.path.join(location, ".svn")):
      apparent_repo_type = "svn"
      break
    else:
      location = location[:location.rfind(os.sep)]

  return apparent_repo_type, location

#end deterimine_repo_type

def rsync(source, dest, options):
  rsync_cmd = ["rsync", "-ar", "--delete"]
  if options.debug_mode:
    rsync_cmd.append("-v")

  if options.source_repo == "git":
    rsync_cmd.append("--exclude=.git")

  rsync_cmd.append(options.source)
  rsync_cmd.append(options.destination)
  run_cmd(rsync_cmd, options)
#end rsync

def create_commit_message(commit_id, commit_log, project_name, project_location):
  eol = os.linesep
  message = "Snapshot of %s from commit %s" % (project_name, commit_id)
  message += eol * 2
  message += "From repository at %s" % project_location
  message += eol * 2
  message += "At commit:" + eol
  message += commit_log
  return message
#end create_commit_message

def find_git_commit_information(options):
  r"""
  >>> class fake_options:
  ...   source="."
  ...   verbose_mode=False
  ...   debug_mode=False
  >>> myoptions = fake_options()
  >>> find_git_commit_information(myoptions)[2:]
  ('sems', 'software.sandia.gov:/git/sems')
  """
  git_log_cmd = ["git", "log", "-1"]
  
  output, error = run_cmd(git_log_cmd, options, options.source)
  
  commit_match = re.match("commit ([0-9a-fA-F]+)", output)
  commit_id = commit_match.group(1)
  commit_log = output
  
  git_remote_cmd = ["git", "remote", "-v"]
  output, error = run_cmd(git_remote_cmd, options, options.source)
  
  remote_match = re.search("origin\s([^ ]*/([^ ]+))", output, re.MULTILINE)
  if not remote_match:
    raise RuntimeError("Could not find origin of repo at %s. Consider using none for source repo type." % (options.source))

  source_location = remote_match.group(1)
  source_name     = remote_match.group(2).strip()
  
  if source_name[-1] == "/":
    source_name = source_name[:-1]

  return commit_id, commit_log, source_name, source_location

#end find_git_commit_information

def do_git_commit(message, options):
  if options.verbose_mode:
    print "Commiting to destination repository."

  git_add_cmd = ["git", "add", "-A"]
  run_cmd(git_add_cmd, options, options.destination)
  
  git_commit_cmd = ["git", "commit", "-m%s" % message]
  run_cmd(git_commit_cmd, options, options.destination)
  
  git_log_cmd = ["git", "log", "--format=%h", "-1"]
  commit_sha1, error = run_cmd(git_log_cmd, options, options.destination)

  print "Commit %s was made to %s." % (commit_sha1.strip(), options.dest_root)
#end do_git_commit

def verify_git_repo_clean(location, options):
  git_status_cmd = ["git", "status", "--porcelain"]
  output, error = run_cmd(git_status_cmd, options, location)
  
  if output != "":
    if options.no_validate_repo == False:
      raise RuntimeError("%s is not clean.%sPlease commit or stash all changes before running snapshot."
        % (location, os.linesep))
    else:
      print "WARNING: %s is not clean. Proceeding anyway." % location
      print "WARNING:   This could lead to differences in the source and destination."
      print "WARNING:   It could also lead to extra files being included in the snapshot commit."

#end verify_git_repo_clean

def main(options):
  if options.verbose_mode:
    print "Snapshotting %s to %s." % (options.source, options.destination)

  if options.source_repo == "git":
    verify_git_repo_clean(options.source, options)
    commit_id, commit_log, repo_name, repo_location = find_git_commit_information(options)
  elif options.source_repo == "none":
    commit_id     = "N/A"
    commit_log    = "Unknown commit from %s snapshotted at: %s" % (options.source, datetime.datetime.now())
    repo_name     = options.source
    repo_location = options.source
    
  commit_message = create_commit_message(commit_id, commit_log, repo_name, repo_location) + os.linesep*2
  
  if options.dest_repo == "git":
    verify_git_repo_clean(options.destination, options)

  rsync(options.source, options.destination, options)
  
  if options.dest_repo == "git":
    do_git_commit(commit_message, options)
  elif options.dest_repo == "none":
    file_name = "snapshot_message.txt"
    message_file = open(file_name, "w")
    message_file.write(commit_message)
    message_file.close()
    cwd = os.getcwd()
    print "No commit done by request. Please use file at:"
    print "%s%sif you wish to commit this to a repo later." % (cwd+"/"+file_name, os.linesep)
  
  
  
  
#end main

if (__name__ == "__main__"):
  if ("--test" in sys.argv):
    doctest.testmod()
    sys.exit(0)

  try:    
    options = parse_cmdline(__doc__)
    main(options)
  except RuntimeError, e:
    print "Error occured:", e
    if "--debug" in sys.argv:
      traceback.print_exc()
    sys.exit(1)
  else:  
    sys.exit(0)



#!/usr/bin/env python3
#
#  Syntax: mkdoc.py [-I<path> ..] [.. a list of header files ..]
#
#  Extract documentation from C++ header files to use it in Python bindings
#

import os
import sys
import platform
import re
import textwrap

from clang import cindex
from clang.cindex import CursorKind
from collections import OrderedDict
from glob import glob
from threading import Thread, Semaphore
from multiprocessing import cpu_count

RECURSE_LIST = [
    CursorKind.TRANSLATION_UNIT,
    CursorKind.NAMESPACE,
    CursorKind.CLASS_DECL,
    CursorKind.STRUCT_DECL,
    CursorKind.ENUM_DECL,
    CursorKind.CLASS_TEMPLATE
]

PRINT_LIST = [
    CursorKind.CLASS_DECL,
    CursorKind.STRUCT_DECL,
    CursorKind.ENUM_DECL,
    CursorKind.ENUM_CONSTANT_DECL,
    CursorKind.CLASS_TEMPLATE,
    CursorKind.FUNCTION_DECL,
    CursorKind.FUNCTION_TEMPLATE,
    CursorKind.CONVERSION_FUNCTION,
    CursorKind.CXX_METHOD,
    CursorKind.CONSTRUCTOR,
    CursorKind.FIELD_DECL
]

PREFIX_BLACKLIST = [
    CursorKind.TRANSLATION_UNIT
]

CPP_OPERATORS = {
    '<=': 'le', '>=': 'ge', '==': 'eq', '!=': 'ne', '[]': 'array',
    '+=': 'iadd', '-=': 'isub', '*=': 'imul', '/=': 'idiv', '%=':
    'imod', '&=': 'iand', '|=': 'ior', '^=': 'ixor', '<<=': 'ilshift',
    '>>=': 'irshift', '++': 'inc', '--': 'dec', '<<': 'lshift', '>>':
    'rshift', '&&': 'land', '||': 'lor', '!': 'lnot', '~': 'bnot',
    '&': 'band', '|': 'bor', '+': 'add', '-': 'sub', '*': 'mul', '/':
    'div', '%': 'mod', '<': 'lt', '>': 'gt', '=': 'assign', '()': 'call'
}

CPP_OPERATORS = OrderedDict(
    sorted(CPP_OPERATORS.items(), key=lambda t: -len(t[0])))

job_count = cpu_count()
job_semaphore = Semaphore(job_count)


class NoFilenamesError(ValueError):
    pass


def d(s):
    return s if isinstance(s, str) else s.decode('utf8')


def sanitize_name(name):
    name = re.sub(r'type-parameter-0-([0-9]+)', r'T\1', name)
    for k, v in CPP_OPERATORS.items():
        name = name.replace('operator%s' % k, 'operator_%s' % v)
    name = re.sub('<.*>', '', name)
    name = ''.join([ch if ch.isalnum() else '_' for ch in name])
    name = re.sub('_$', '', re.sub('_+', '_', name))
    return '__doc_' + name


def process_comment(comment):
    result = ''

    # Remove C++ comment syntax
    leading_spaces = float('inf')
    for s in comment.expandtabs(tabsize=4).splitlines():
        s = s.strip()
        if s.startswith('/*'):
            s = s[2:].lstrip('*')
        elif s.endswith('*/'):
            s = s[:-2].rstrip('*')
        elif s.startswith('///'):
            s = s[3:]
        if s.startswith('*'):
            s = s[1:]
        if len(s) > 0:
            leading_spaces = min(leading_spaces, len(s) - len(s.lstrip()))
        result += s + '\n'

    if leading_spaces != float('inf'):
        result2 = ""
        for s in result.splitlines():
            result2 += s[leading_spaces:] + '\n'
        result = result2

    # Doxygen tags
    cpp_group = '([\w:]+)'
    param_group = '([\[\w:\]]+)'

    s = result
    s = re.sub(r'\\c\s+%s' % cpp_group, r'``\1``', s)
    s = re.sub(r'\\a\s+%s' % cpp_group, r'*\1*', s)
    s = re.sub(r'\\e\s+%s' % cpp_group, r'*\1*', s)
    s = re.sub(r'\\em\s+%s' % cpp_group, r'*\1*', s)
    s = re.sub(r'\\b\s+%s' % cpp_group, r'**\1**', s)
    s = re.sub(r'\\ingroup\s+%s' % cpp_group, r'', s)
    s = re.sub(r'\\param%s?\s+%s' % (param_group, cpp_group),
               r'\n\n$Parameter ``\2``:\n\n', s)
    s = re.sub(r'\\tparam%s?\s+%s' % (param_group, cpp_group),
               r'\n\n$Template parameter ``\2``:\n\n', s)

    for in_, out_ in {
        'return': 'Returns',
        'author': 'Author',
        'authors': 'Authors',
        'copyright': 'Copyright',
        'date': 'Date',
        'remark': 'Remark',
        'sa': 'See also',
        'see': 'See also',
        'extends': 'Extends',
        'throw': 'Throws',
        'throws': 'Throws'
    }.items():
        s = re.sub(r'\\%s\s*' % in_, r'\n\n$%s:\n\n' % out_, s)

    s = re.sub(r'\\details\s*', r'\n\n', s)
    s = re.sub(r'\\brief\s*', r'', s)
    s = re.sub(r'\\short\s*', r'', s)
    s = re.sub(r'\\ref\s*', r'', s)

    s = re.sub(r'\\code\s?(.*?)\s?\\endcode',
               r"```\n\1\n```\n", s, flags=re.DOTALL)

    # HTML/TeX tags
    s = re.sub(r'<tt>(.*?)</tt>', r'``\1``', s, flags=re.DOTALL)
    s = re.sub(r'<pre>(.*?)</pre>', r"```\n\1\n```\n", s, flags=re.DOTALL)
    s = re.sub(r'<em>(.*?)</em>', r'*\1*', s, flags=re.DOTALL)
    s = re.sub(r'<b>(.*?)</b>', r'**\1**', s, flags=re.DOTALL)
    s = re.sub(r'\\f\$(.*?)\\f\$', r'$\1$', s, flags=re.DOTALL)
    s = re.sub(r'<li>', r'\n\n* ', s)
    s = re.sub(r'</?ul>', r'', s)
    s = re.sub(r'</li>', r'\n\n', s)

    s = s.replace('``true``', '``True``')
    s = s.replace('``false``', '``False``')

    # Re-flow text
    wrapper = textwrap.TextWrapper()
    wrapper.expand_tabs = True
    wrapper.replace_whitespace = True
    wrapper.drop_whitespace = True
    wrapper.width = 70
    wrapper.initial_indent = wrapper.subsequent_indent = ''

    result = ''
    in_code_segment = False
    for x in re.split(r'(```)', s):
        if x == '```':
            if not in_code_segment:
                result += '```\n'
            else:
                result += '\n```\n\n'
            in_code_segment = not in_code_segment
        elif in_code_segment:
            result += x.strip()
        else:
            for y in re.split(r'(?: *\n *){2,}', x):
                wrapped = wrapper.fill(re.sub(r'\s+', ' ', y).strip())
                if len(wrapped) > 0 and wrapped[0] == '$':
                    result += wrapped[1:] + '\n'
                    wrapper.initial_indent = \
                        wrapper.subsequent_indent = ' ' * 4
                else:
                    if len(wrapped) > 0:
                        result += wrapped + '\n\n'
                    wrapper.initial_indent = wrapper.subsequent_indent = ''
    return result.rstrip().lstrip('\n')


def extract(filename, node, prefix, output):
    if not (node.location.file is None or
            os.path.samefile(d(node.location.file.name), filename)):
        return 0
    if node.kind in RECURSE_LIST:
        sub_prefix = prefix
        if node.kind not in PREFIX_BLACKLIST:
            if len(sub_prefix) > 0:
                sub_prefix += '_'
            sub_prefix += d(node.spelling)
        for i in node.get_children():
            extract(filename, i, sub_prefix, output)
    if node.kind in PRINT_LIST:
        comment = d(node.raw_comment) if node.raw_comment is not None else ''
        comment = process_comment(comment)
        sub_prefix = prefix
        if len(sub_prefix) > 0:
            sub_prefix += '_'
        if len(node.spelling) > 0:
            name = sanitize_name(sub_prefix + d(node.spelling))
            output.append((name, filename, comment))


class ExtractionThread(Thread):
    def __init__(self, filename, parameters, output):
        Thread.__init__(self)
        self.filename = filename
        self.parameters = parameters
        self.output = output
        job_semaphore.acquire()

    def run(self):
        print('Processing "%s" ..' % self.filename, file=sys.stderr)
        try:
            index = cindex.Index(
                cindex.conf.lib.clang_createIndex(False, True))
            tu = index.parse(self.filename, self.parameters)
            extract(self.filename, tu.cursor, '', self.output)
        finally:
            job_semaphore.release()


def read_args(args):
    parameters = []
    filenames = []
    if "-x" not in args:
        parameters.extend(['-x', 'c++'])
    if not any(it.startswith("-std=") for it in args):
        parameters.append('-std=c++11')

    if platform.system() == 'Darwin':
        dev_path = '/Applications/Xcode.app/Contents/Developer/'
        lib_dir = dev_path + 'Toolchains/XcodeDefault.xctoolchain/usr/lib/'
        sdk_dir = dev_path + 'Platforms/MacOSX.platform/Developer/SDKs'
        libclang = lib_dir + 'libclang.dylib'

        if os.path.exists(libclang):
            cindex.Config.set_library_path(os.path.dirname(libclang))

        if os.path.exists(sdk_dir):
            sysroot_dir = os.path.join(sdk_dir, next(os.walk(sdk_dir))[1][0])
            parameters.append('-isysroot')
            parameters.append(sysroot_dir)
    elif platform.system() == 'Linux':
        # cython.util.find_library does not find `libclang` for all clang
        # versions and distributions. LLVM switched to a monolithical setup
        # that includes everything under /usr/lib/llvm{version_number}/
        # We therefore glob for the library and select the highest version
        library_file = sorted(glob("/usr/lib/llvm-*/lib/libclang.so"), reverse=True)[0]
        cindex.Config.set_library_file(library_file)

        # clang doesn't find its own base includes by default on Linux,
        # but different distros install them in different paths.
        # Try to autodetect, preferring the highest numbered version.
        def clang_folder_version(d):
            return [int(ver) for ver in re.findall(r'(?<!lib)(?<!\d)\d+', d)]
        clang_include_dir = max((
            path
            for libdir in ['lib64', 'lib', 'lib32']
            for path in glob('/usr/%s/clang/*/include' % libdir)
            if os.path.isdir(path)
        ), default=None, key=clang_folder_version)
        if clang_include_dir:
            parameters.extend(['-isystem', clang_include_dir])

    for item in args:
        if item.startswith('-'):
            parameters.append(item)
        else:
            filenames.append(item)

    if len(filenames) == 0:
        raise NoFilenamesError("args parameter did not contain any filenames")

    return parameters, filenames


def extract_all(args):
    parameters, filenames = read_args(args)
    output = []
    for filename in filenames:
        thr = ExtractionThread(filename, parameters, output)
        thr.start()

    print('Waiting for jobs to finish ..', file=sys.stderr)
    for i in range(job_count):
        job_semaphore.acquire()

    return output


def write_header(comments, out_file=sys.stdout):
    print('''/*
  This file contains docstrings for the Python bindings.
  Do not edit! These were automatically extracted by mkdoc.py
 */

#define __EXPAND(x)                                      x
#define __COUNT(_1, _2, _3, _4, _5, _6, _7, COUNT, ...)  COUNT
#define __VA_SIZE(...)                                   __EXPAND(__COUNT(__VA_ARGS__, 7, 6, 5, 4, 3, 2, 1))
#define __CAT1(a, b)                                     a ## b
#define __CAT2(a, b)                                     __CAT1(a, b)
#define __DOC1(n1)                                       __doc_##n1
#define __DOC2(n1, n2)                                   __doc_##n1##_##n2
#define __DOC3(n1, n2, n3)                               __doc_##n1##_##n2##_##n3
#define __DOC4(n1, n2, n3, n4)                           __doc_##n1##_##n2##_##n3##_##n4
#define __DOC5(n1, n2, n3, n4, n5)                       __doc_##n1##_##n2##_##n3##_##n4##_##n5
#define __DOC6(n1, n2, n3, n4, n5, n6)                   __doc_##n1##_##n2##_##n3##_##n4##_##n5##_##n6
#define __DOC7(n1, n2, n3, n4, n5, n6, n7)               __doc_##n1##_##n2##_##n3##_##n4##_##n5##_##n6##_##n7
#define DOC(...)                                         __EXPAND(__EXPAND(__CAT2(__DOC, __VA_SIZE(__VA_ARGS__)))(__VA_ARGS__))

#if defined(__GNUG__)
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-variable"
#endif
''', file=out_file)


    name_ctr = 1
    name_prev = None
    for name, _, comment in list(sorted(comments, key=lambda x: (x[0], x[1]))):
        if name == name_prev:
            name_ctr += 1
            name = name + "_%i" % name_ctr
        else:
            name_prev = name
            name_ctr = 1
        print('\nstatic const char *%s =%sR"doc(%s)doc";' %
              (name, '\n' if '\n' in comment else ' ', comment), file=out_file)

    print('''
#if defined(__GNUG__)
#pragma GCC diagnostic pop
#endif
''', file=out_file)


def mkdoc(args):
    args = list(args)
    out_path = None
    for idx, arg in enumerate(args):
        if arg.startswith("-o"):
            args.remove(arg)
            try:
                out_path = arg[2:] or args.pop(idx)
            except IndexError:
                print("-o flag requires an argument")
                exit(-1)
            break

    comments = extract_all(args)

    if out_path:
        try:
            with open(out_path, 'w') as out_file:
                write_header(comments, out_file)
        except:
            # In the event of an error, don't leave a partially-written
            # output file.
            try:
                os.unlink(out_path)
            except:
                pass
            raise
    else:
        write_header(comments)


if __name__ == '__main__':
    try:
        mkdoc(sys.argv[1:])
    except NoFilenamesError:
        print('Syntax: %s [.. a list of header files ..]' % sys.argv[0])
        exit(-1)



from unittest import TestCase
import numpy as np
import math
import random
import pycompadre

kokkos_obj=pycompadre.KokkosParser()

# function used to generate sample data
def exact(coord,order,dimension):
    x = coord[0]
    y = coord[1] if (dimension>1) else 0
    z = coord[2] if (dimension>2) else 0
    if order==1:
        return 1 + x + y + z
    elif order==2:
        return 1 + x + y + z + x*x + x*y + x*z + y*y + y*z + z*z
    elif order==3:
        return 1 + x + y + z + x*x + x*y + x*z + y*y + y*z + z*z + x*x*x + x*x*y + x*x*z + x*y*y + x*y*z + x*z*z + y*y*y + y*y*z + y*z*z + z*z*z

# function used to get analytic gradient
def grad_exact(coord,component,order,dimension):
    x = coord[0]
    y = coord[1] if (dimension>1) else 0
    z = coord[2] if (dimension>2) else 0
    if (component==0):
        if order==1:
            return 1
        elif order==2:
            return 1 + 2*x + y + z
        elif order==3:
            return 1 + 2*x + y + z + 3*x*x + 2*x*y + 2*x*z + y*y + y*z + z*z
    elif (component==1):
        if order==1:
            return 1
        elif order==2:
            return 1 + x + 2*y + z
        elif order==3:
            return 1 + x + 2*y + z + x*x + x*2*y + x*z + 3*y*y + 2*y*z + z*z
    elif (component==2):
        if order==1:
            return 1
        elif order==2:
            return 1 + x + y + 2*z
        elif order==3:
            return 1 + x + y + 2*z + x*x + x*y + x*2*z + y*y + y*2*z + 3*z*z

def remap(polyOrder,dimension,additional_sites=False):

    minND = [[10,20,30],[10,20,100],[30,30,60]]
    ND = minND[dimension-1][polyOrder-1]

    random.seed(1234) # for consistent results

    dimensions = dimension

    # initialize 3rd order reconstruction using 2nd order basis in 3D (GMLS)
    gmls_obj=pycompadre.GMLS(polyOrder, dimensions, "QR", "STANDARD")
    gmls_obj.setWeightingParameter(4)
    gmls_obj.setWeightingType("power")

    NT = 10 # Targets
    nx, ny, nz = (ND, ND, ND)

    xmax = 1
    if (dimension>1): ymax = 1
    if (dimension>2): zmax = 1
    xmin = -xmax
    if (dimension>1): ymin = -ymax
    if (dimension>2): zmin = -zmax

    dx = np.linspace(xmin, xmax, nx)
    if (dimension>1): dy = np.linspace(ymin, ymax, ny)
    if (dimension>2): dz = np.linspace(zmin, zmax, nz)

    N=1
    for i in range(dimension):
        N*=ND

    # target sites
    target_sites = []
    for i in range(NT):
        if (dimension==1):
            target_sites.append([random.uniform(xmin, xmax)])
        elif (dimension==2):
            target_sites.append([random.uniform(xmin, xmax), random.uniform(ymin, ymax)])
        elif (dimension==3):
            target_sites.append([random.uniform(xmin, xmax), random.uniform(ymin, ymax), random.uniform(zmin, zmax)])
    target_sites = np.array(target_sites, dtype='d')

    # source sites
    t_sites = list()
    for i in range(ND):
        if (dimension==1):
            t_sites.append([dx[i],])
        else:
            for j in range(ND):
                if (dimension==2):
                    t_sites.append([dx[i],dy[j]])
                else:
                    for k in range(ND):
                        t_sites.append([dx[i],dy[j],dz[k]])
    source_sites = np.array(t_sites, dtype=np.dtype('d'))

    # neighbor search
    epsilon_multiplier = 1.5
    gmls_helper = pycompadre.ParticleHelper(gmls_obj)
    gmls_helper.generateKDTree(source_sites)
    gmls_helper.generateNeighborListsFromKNNSearchAndSet(target_sites, polyOrder, dimensions, epsilon_multiplier)

    # set data in gmls object
    gmls_helper.setSourceSites(source_sites)

    # used in combination with polynomial coefficients
    epsilons = gmls_helper.getWindowSizes()

    gmls_obj.addTargets(pycompadre.TargetOperation.ScalarPointEvaluation)
    gmls_obj.addTargets(pycompadre.TargetOperation.PartialXOfScalarPointEvaluation)
    (dimensions>1) and gmls_obj.addTargets(pycompadre.TargetOperation.PartialYOfScalarPointEvaluation)
    (dimensions>2) and gmls_obj.addTargets(pycompadre.TargetOperation.PartialYOfScalarPointEvaluation)

    # add additional evaluation sites (if specified)
    if additional_sites:
        additional_sites_indices = np.zeros(shape=(NT,5), dtype='i4')
        additional_site_coordinates = np.zeros(shape=(4*NT,dimension), dtype='f8')
        additional_sites_indices[:,0] = 4
        inds = np.arange(0, 4*NT)
        inds = np.reshape(inds, newshape=(-1,4))
        additional_sites_indices[:,1::] = inds
        h = np.linalg.norm(source_sites[0,:]-source_sites[1,:])
        for i in range(NT):
            for j in range(4):
                for k in range(dimension):
                    additional_site_coordinates[i*4+j,k] = target_sites[i,k] + random.uniform(-h, h)

        gmls_helper.setAdditionalEvaluationSitesData(additional_sites_indices, additional_site_coordinates)

    # generate stencil with number of batches set to 1, and keeping coefficients (not necessary)
    gmls_obj.generateAlphas(1, True)

    # create sample data at source sites
    data_vector = []
    for i in range(N):
        data_vector.append(exact(source_sites[i], polyOrder, dimension))
    # use rank 2 array and only insert into one column to test
    # whether layouts are being properly propagated into pycompadre
    new_data_vector = np.zeros(shape=(len(data_vector), 3), dtype='f8')
    new_data_vector[:,1] = np.array(data_vector, dtype=np.dtype('d'))

    # apply stencil to sample data for all targets
    computed_answer = gmls_helper.applyStencil(new_data_vector[:,1], pycompadre.TargetOperation.ScalarPointEvaluation)

    l2_error = 0
    for i in range(NT):
        l2_error += np.power(abs(computed_answer[i] - exact(target_sites[i],polyOrder,dimension)),2)
    l2_error = math.sqrt(l2_error/float(NT))

    additional_sites_l2_error = 0.0
    if additional_sites:
        nl = gmls_helper.getNeighborLists()
        # test min/max num neighbors computation works
        nl.computeMinNumNeighbors()
        nl.computeMaxNumNeighbors()
        n_min = nl.getMinNumNeighbors()
        n_max = nl.getMaxNumNeighbors()
        for i in range(NT):
            for j in range(4):
                computed_answer = 0.0
                for k in range(nl.getNumberOfNeighbors(i)):
                    computed_answer += gmls_obj.getSolutionSet().getAlpha(pycompadre.TargetOperation.ScalarPointEvaluation, 
                                                            i, 0, 0, k, 0, 0, j+1)*data_vector[nl.getNeighbor(i,k)]
                additional_sites_l2_error += \
                    np.power(abs(computed_answer - exact(additional_site_coordinates[i*4+j],polyOrder,dimension)),2)

    # retrieve additional sites neighbor lists just to make sure it works
    a_nl = gmls_helper.getAdditionalEvaluationIndices()

    # get polynomial coefficients
    polynomial_coefficients = gmls_helper.getPolynomialCoefficients(data_vector)

    # alternative way to compute h1 semi norm
    # could remap using the gradient operator, but instead this uses calculated polynomial coefficients and applies
    # the action of the gradient target operation on the polynomial basis at the target sites
    # this serves as a test for getting accurate calculation and retrieval of polynomial coefficients using
    # the python interface
    h1_seminorm_error = 0
    for i in range(NT):
        h1_seminorm_error += np.power(abs(1./epsilons[i]*polynomial_coefficients[i,1] - grad_exact(target_sites[i], 0, polyOrder, dimension)),2)
        if (dimension>1): h1_seminorm_error += np.power(abs(1./epsilons[i]*polynomial_coefficients[i,2] - grad_exact(target_sites[i], 1, polyOrder, dimension)),2)
        if (dimension>2): h1_seminorm_error += np.power(abs(1./epsilons[i]*polynomial_coefficients[i,3] - grad_exact(target_sites[i], 2, polyOrder, dimension)),2)
    h1_seminorm_error = math.sqrt(h1_seminorm_error/float(NT))

    if additional_sites:
        return l2_error, h1_seminorm_error, additional_sites_l2_error
    return l2_error, h1_seminorm_error


class TestPycompadre(TestCase):
    def test_1d_order1(self):
        l2,h1=remap(1,1)
        self.assertTrue(l2<1e-13 and h1<1e-13)
    def test_1d_order2(self):
        l2,h1=remap(2,1)
        self.assertTrue(l2<1e-13 and h1<1e-13)
    def test_1d_order3(self):
        l2,h1=remap(3,1)
        self.assertTrue(l2<1e-13 and h1<1e-13)

    def test_2d_order1(self):
        l2,h1=remap(1,2)
        self.assertTrue(l2<1e-13 and h1<1e-13)
    def test_2d_order2(self):
        l2,h1=remap(2,2)
        self.assertTrue(l2<1e-13 and h1<1e-13)
    def test_2d_order3(self):
        l2,h1=remap(3,2)
        self.assertTrue(l2<1e-13 and h1<1e-13)

    def test_3d_order1(self):
        l2,h1=remap(1,3)
        self.assertTrue(l2<1e-13 and h1<1e-13)
    def test_3d_order2(self):
        l2,h1=remap(2,3)
        self.assertTrue(l2<1e-13 and h1<1e-13)
    def test_3d_order3(self):
        l2,h1=remap(3,3)
        self.assertTrue(l2<1e-13 and h1<1e-13)

    def test_additional_sites(self):
        l2,h1,l2a=remap(2,1,True)
        self.assertTrue(l2<1e-13 and h1<1e-13 and l2a<1e-13)
        l2,h1,l2a=remap(1,2,True)
        self.assertTrue(l2<1e-13 and h1<1e-13 and l2a<1e-13)
        l2,h1,l2a=remap(3,3,True)
        self.assertTrue(l2<1e-13 and h1<1e-13 and l2a<1e-13)

#tc = TestPycompadre()
#tc.test_additional_sites()
#tc.test_3d_order1()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
import sys
import os
import pycompadre
#tls_cacert = "/Users/pakuber/sandia_root_ca.cer"
tls_verify = False


# -- Project information -----------------------------------------------------

project = 'PyCOMPADRE v%s'%pycompadre.__version__
copyright = '2021, National Technology & Engineering Solutions of Sandia, LLC (NTESS)'
author = 'Paul Kuberry'

# The full version, including alpha/beta/rc tags
release = pycompadre.__version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
]
autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['pybind11/*','build']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

#latex_documents = [
#    (master_doc, 'python_example.tex', u'python_example Documentation',
#     u'Sylvain Corlay', 'manual'),
#]
#pygments_style = 'sphinx'
#man_pages = [
#    (master_doc, 'python_example', u'python_example Documentation',
#     [author], 1)
#]
#texinfo_documents = [
#    (master_doc, 'python_example', u'python_example Documentation',
#     author, 'python_example', 'One line description of project.',
#     'Miscellaneous'),
#]
intersphinx_mapping = {'https://docs.python.org/': None}



import pycompadre
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button, RadioButtons, CheckButtons

''' Interactive example of pycompadre in one dimension

By changing the evaluation location, users can see the local polynomial approximation
and observe how it changes as it is drug left or right.

Additional, users can get a feel for the effect of various kernels, polynomial orders,
and epsilon multiplier choices.
'''

# initialize Kokkos
kp = pycompadre.KokkosParser()

# initialize parameters
solver_type = 'QR'
polynomial_order = 2
input_dimensions = 1
epsilon_multiplier = 1.6
global weighting_type, function, num_data_points
weighting_type = 'power'
function = lambda x: pow(x,2)
num_data_points = 10
x = np.linspace(0,4,num_data_points)
y = function(x)
x_pred = np.linspace(0,4,2000)
extra_sites_coords = np.atleast_2d(np.linspace(0,4,200)).T

# get GMLS approximate at all x_pred, as well as reconstruction about attempt_center_about_coord
def approximate(solver_type, porder, wpower0, wpower1, wtype, epsilon_multiplier, attempt_center_about_coord):
    gmls_obj=pycompadre.GMLS(porder, input_dimensions, solver_type, "STANDARD")
    gmls_obj.setWeightingParameter(wpower0,0)
    gmls_obj.setWeightingParameter(wpower1,1)
    gmls_obj.setWeightingType(wtype)
    gmls_helper = pycompadre.ParticleHelper(gmls_obj)
    gmls_helper.generateKDTree(np.atleast_2d(x).T)
    gmls_obj.addTargets(pycompadre.TargetOperation.ScalarPointEvaluation)
    
    gmls_helper.generateNeighborListsFromKNNSearchAndSet(np.atleast_2d(x_pred).T, porder, input_dimensions, epsilon_multiplier, 0.0, False, True)
    
    center_about_idx   = (np.abs(x_pred - attempt_center_about_coord)).argmin()
    center_about_coord = x_pred[center_about_idx]
    extra_sites_idx    = np.zeros(shape=(len(x_pred),len(extra_sites_coords)+1), dtype='i4')
    extra_sites_idx[center_about_idx,0] = len(extra_sites_coords)
    extra_sites_idx[center_about_idx,1:] = np.arange(len(extra_sites_coords))
    gmls_helper.setAdditionalEvaluationSitesData(extra_sites_idx, extra_sites_coords)
    
    gmls_obj.generateAlphas(1, False)
    
    # helper function for applying of alphas
    y_pred = gmls_helper.applyStencil(np.atleast_2d(y).T, pycompadre.TargetOperation.ScalarPointEvaluation)
    
    # manual applying of alphas
    nl = gmls_helper.getNeighborLists()
    for i in range(len(x_pred)):
        if (i==center_about_idx):
            computed_answer = np.zeros(shape=(len(extra_sites_coords),), dtype='f8')
            colors = len(x)*['black']
            for j in range(extra_sites_idx[i,0]):
                computed_answer[j] = 0.0
                for k in range(nl.getNumberOfNeighbors(i)):
                    colors[nl.getNeighbor(i,k)] = 'red'
                    computed_answer[j] += gmls_obj.getSolutionSet().getAlpha(pycompadre.TargetOperation.ScalarPointEvaluation, 
                                          i, 0, 0, k, 0, 0, j+1)*y[nl.getNeighbor(i,k)]
    center_about_extra_idx   = (np.abs(extra_sites_coords - center_about_coord)).argmin()
    center_about_extra_coord = extra_sites_coords[center_about_extra_idx]
    del nl
    del gmls_obj
    del gmls_helper
    return (y_pred, computed_answer, center_about_extra_idx, center_about_extra_coord, colors)

# get initial data for plotting
y_pred, computed_answer, center_about_extra_idx, center_about_extra_coord, colors = approximate(solver_type, polynomial_order, 2, 1, 'power', epsilon_multiplier, 1.2)
# plot initial data
fig, ax = plt.subplots()
plt.subplots_adjust(left=0.25, bottom=0.25)
p = ax.scatter(x, y, c='#000000', marker='D', zorder=2)
p.set_color(colors)
d, = plt.plot(x_pred, y_pred, c='#0000FF', linewidth=4, zorder=1)
l, = plt.plot(extra_sites_coords, computed_answer, c='#00FF00', lw=2, zorder=3)
s = plt.scatter([center_about_extra_coord,], computed_answer[center_about_extra_idx], c='#00FF00', marker='o', zorder=4, edgecolor='black', s=50)
ax.set(xlabel='x', ylabel='',
       title='GMLS Approximation')
ax.grid()
ax.margins(x=0)
axcolor = 'lightgoldenrodyellow'
plt.legend((p,d,l,s), ('Data', 'GMLS', 'Local Polynomial Fit', 'Target Site'))

# axes for sliders and radio buttons
ax_location_check = plt.axes([0.10, 0.068, 0.15, 0.15], facecolor='white', frameon=False)
ax_location = plt.axes([0.25, 0.125, 0.65, 0.03], facecolor='white')
ax_weighting_power_0 = plt.axes([0.6175, 0.025, 0.075, 0.03], facecolor=axcolor)
ax_weighting_power_1 = plt.axes([0.825, 0.025, 0.075, 0.03], facecolor=axcolor)
ax_num_data_points = plt.axes([0.25, 0.025, 0.2, 0.03], facecolor=axcolor)
ax_epsilon = plt.axes([0.7, 0.075, 0.2, 0.03])
ax_polynomial_order = plt.axes([0.25, 0.075, 0.2, 0.03])
ax_weighting_type = plt.axes([0.015, 0.20, 0.15, 0.20], facecolor=axcolor)
ax_func_type = plt.axes([0.015, 0.40, 0.15, 0.15], facecolor=axcolor)
ax_solver_type = plt.axes([0.015, 0.55, 0.15, 0.20], facecolor=axcolor)

# sliders
sl_location_check = CheckButtons(ax_location_check, ["",], [True,])
delta_f = 4.0/200
sl_location = Slider(ax_location, 'Location', valmin=0.0, valmax=4.0, valinit=0.0, valstep=delta_f, color=None, initcolor='black')
sl_location.set_val(1.2)
sl_weighting_power_0 = Slider(ax_weighting_power_0, 'Wgt. P(0)', valmin=0, valmax=6, valinit=2, valstep=1)
sl_weighting_power_1 = Slider(ax_weighting_power_1, 'Wgt. P(1)', valmin=0, valmax=6, valinit=1, valstep=1)
sl_num_data_points = Slider(ax_num_data_points, 'Number of Points', valmin=6, valmax=50, valinit=10, valstep=1)
sl_epsilon = Slider(ax_epsilon, 'Epsilon Multipler', valmin=1.0001, valmax=5.0, valinit=1.6, valstep=.1)
sl_polynomial_order = Slider(ax_polynomial_order, 'Polynomial Order', valmin=0, valmax=6, valinit=2, valstep=1)

#radios
rad_weighting_type = RadioButtons(ax_weighting_type, ('Power', 'Cubic Spl.', 'Cosine', 'Gaussian', 'Sigmoid'), active=0)
rad_func_type = RadioButtons(ax_func_type, ('sin(x)', 'x*sin(20x)', 'x^2'), active=2)
rad_solver_type = RadioButtons(ax_solver_type, ('QR', 'LU'), active=0)

def update(val, relim=False):
    sl_location.valinit=2.0
    global weighting_type
    y_pred, computed_answer, center_about_extra_idx, center_about_extra_coord, colors = approximate(rad_solver_type.value_selected, sl_polynomial_order.val, sl_weighting_power_0.val, sl_weighting_power_1.val, weighting_type, sl_epsilon.val, sl_location.val)
    l.set_ydata(computed_answer)
    d.set_ydata(y_pred)
    s.set_offsets([center_about_extra_coord, computed_answer[center_about_extra_idx]])
    if sl_location_check.get_status()[0]:
        p.set_color(colors)
    if relim:
        ax.ignore_existing_data_limits = True
        ax.update_datalim(p.get_datalim(ax.transData))
        ax.autoscale(True)
        ax.autoscale_view()
        ax.autoscale(False)
    fig.canvas.draw_idle()
# register objects using update
sl_location.on_changed(update)
sl_weighting_power_0.on_changed(update)
sl_weighting_power_1.on_changed(update)
sl_epsilon.on_changed(update)
sl_polynomial_order.on_changed(update)
rad_solver_type.on_clicked(update)

# called from changefunc
def updatepoints(val):
    global num_data_points, x, y 
    num_data_points = sl_num_data_points.val
    x = np.resize(x, new_shape=(num_data_points,))
    y = np.resize(y, new_shape=(num_data_points,))
    x[:] = np.linspace(0,4,num_data_points)

def weighting_type_update(label):
    weighting_type_dict = {'Power': 'power', 'Cubic Spl.': 'cubicspline', 'Cosine': 'cosine', 'Gaussian': 'gaussian', 'Sigmoid': 'sigmoid'}
    global weighting_type
    weighting_type = weighting_type_dict[label]
    if (label=='Power'):
        ax_weighting_power_0.set_visible(True)
        ax_weighting_power_1.set_visible(True)
    elif (label=='Cubic Spl.'):
        ax_weighting_power_0.set_visible(False)
        ax_weighting_power_1.set_visible(False)
    elif (label=='Cosine'):
        ax_weighting_power_0.set_visible(False)
        ax_weighting_power_1.set_visible(False)
    elif (label=='Gaussian'):
        ax_weighting_power_0.set_visible(True)
        ax_weighting_power_1.set_visible(False)
    elif (label=='Sigmoid'):
        ax_weighting_power_0.set_visible(True)
        ax_weighting_power_1.set_visible(True)
    update(0)
# register objects using weighting_type_update
rad_weighting_type.on_clicked(weighting_type_update)

def changefunc(label):
    updatepoints(0)
    global weighting_type
    global function
    if type(label)==str:
        func_type_dict = {'sin(x)': lambda x: np.sin(x), 'x*sin(20x)': lambda x: x*np.sin(20*x), 'x^2' : lambda x: pow(x,2)}
        function = func_type_dict[label]
    y[:] = function(x)
    p.set_offsets(np.vstack((x,y)).T)
    update(0, True)
    fig.canvas.draw_idle()
# register objects using changefunc
rad_func_type.on_clicked(changefunc)
sl_num_data_points.on_changed(changefunc)

def changelocationviz(label):
    l.set_visible(sl_location_check.get_status()[0])
    fig.canvas.draw_idle()
    if sl_location_check.get_status()[0]:
        update(0)
    else:
        p.set_color('black')
sl_location_check.on_clicked(changelocationviz)

plt.show()
del kp



import time
import pycompadre
import numpy as np
from mpl_toolkits import mplot3d
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button, RadioButtons, CheckButtons

''' Interactive example of pycompadre in two dimensions

By changing the evaluation location, users can see the local polynomial approximation
and observe how it changes as it is drug left or right.

Additional, users can get a feel for the effect of various kernels, polynomial orders,
and epsilon multiplier choices.
'''

PRINT_SETUP_TIME = False 
PRINT_SOLVE_TIME = False
PRINT_PLOT_TIME  = False 

# initialize Kokkos
kp = pycompadre.KokkosParser()

# initialize parameters
tic = time.perf_counter()
polynomial_order = 2
input_dimensions = 2
epsilon_multiplier = 1.6
global weighting_type, function, num_data_points
weighting_type = 'power'
function = lambda x, y: pow(x,2)*pow(y,2)
num_data_points = 10
global x, y, X, Y, XY_ravel, Z, Z_ravel
x = np.linspace(0,4,num_data_points)
y = np.linspace(0,4,num_data_points)
X, Y = np.meshgrid(x,y)
XY_ravel = np.vstack((X.ravel(order = 'C'),Y.ravel(order = 'C'))).T
Z = function(X,Y)
Z_ravel = Z.ravel(order='C')
x_pred = np.linspace(0,4,200)
y_pred = np.linspace(0,4,200)
X_pred, Y_pred = np.meshgrid(x_pred,y_pred)
XY_pred_ravel = np.vstack((X_pred.ravel(order = 'C'),Y_pred.ravel(order = 'C'))).T
extra_sites_coords_X = np.copy(X_pred)
extra_sites_coords_Y = np.copy(Y_pred)
toc = time.perf_counter()
PRINT_SETUP_TIME and print("Setup data in %0.6f seconds"%(toc-tic,))
    
# get GMLS approximate at all x_pred, as well as reconstruction about attempt_center_about_coord
def approximate(porder, wpower, wtype, epsilon_multiplier, attempt_center_about_coord):
    tic = time.perf_counter()
    gmls_obj=pycompadre.GMLS(porder, input_dimensions, "QR", "STANDARD")
    gmls_obj.setWeightingParameter(wpower)
    gmls_obj.setWeightingType(wtype)
    gmls_helper = pycompadre.ParticleHelper(gmls_obj)

    gmls_helper.generateKDTree(XY_ravel)
    gmls_obj.addTargets(pycompadre.TargetOperation.ScalarPointEvaluation)
    
    gmls_helper.generateNeighborListsFromKNNSearchAndSet(XY_pred_ravel, porder, input_dimensions, epsilon_multiplier)
    gmls_obj.generateAlphas(1, False)
    
    # helper function for applying of alphas
    Z_pred = gmls_helper.applyStencil(Z_ravel, pycompadre.TargetOperation.ScalarPointEvaluation)

    center_about_idx   = np.sum(np.abs(XY_pred_ravel - attempt_center_about_coord), axis=1).argmin()
    center_about_coord = XY_pred_ravel[center_about_idx]
    extra_sites_coords = np.copy(XY_pred_ravel)

    gmls_obj_2=pycompadre.GMLS(porder, input_dimensions, "QR", "STANDARD")
    gmls_obj_2.setWeightingParameter(wpower)
    gmls_obj_2.setWeightingType(wtype)
    gmls_helper_2 = pycompadre.ParticleHelper(gmls_obj_2)
    gmls_helper_2.generateKDTree(XY_ravel)
    gmls_obj_2.addTargets(pycompadre.TargetOperation.ScalarPointEvaluation)
    gmls_helper_2.generateNeighborListsFromKNNSearchAndSet(np.atleast_2d(XY_pred_ravel[center_about_idx]), porder, input_dimensions, epsilon_multiplier)

    extra_sites_idx = np.zeros(shape=(1,extra_sites_coords.shape[0]+1), dtype='i4')
    extra_sites_idx[0,0] = extra_sites_coords.shape[0]
    extra_sites_idx[0,1:] = np.arange(extra_sites_coords.shape[0])
    gmls_helper_2.setAdditionalEvaluationSitesData(extra_sites_idx, extra_sites_coords)

    gmls_obj_2.generateAlphas(1, False)
    
    # manual applying of alphas
    nl = gmls_helper_2.getNeighborLists()
    computed_answer = np.zeros(shape=(len(extra_sites_coords),), dtype='f8')
    sf = pycompadre.SamplingFunctional['PointSample']

    colors = len(XY_ravel)*['black']
    for k in range(nl.getNumberOfNeighbors(0)):
        colors[nl.getNeighbor(0,k)] = 'red'
    for j in range(extra_sites_idx[0,0]):
        computed_answer[j] = gmls_helper_2.applyStencilSingleTarget(Z_ravel, pycompadre.TargetOperation.ScalarPointEvaluation, sf, j+1)

    center_about_extra_idx   = np.sum(np.abs(extra_sites_coords - center_about_coord), axis=1).argmin()
    center_about_extra_coord = extra_sites_coords[center_about_extra_idx]
    del nl
    del gmls_obj
    del gmls_obj_2
    del gmls_helper
    del gmls_helper_2
    toc = time.perf_counter()
    PRINT_SOLVE_TIME and print("Solve GMLS in %0.6f seconds"%(toc-tic,))
    return (np.reshape(Z_pred, newshape=(len(x_pred), len(y_pred))), np.reshape(computed_answer, newshape=(len(x_pred), len(y_pred))), center_about_extra_idx, center_about_extra_coord, colors)
    
# get initial data for plotting
Z_pred, computed_answer, center_about_extra_idx, center_about_extra_coord, colors = approximate(polynomial_order, 3, 'power', epsilon_multiplier, np.atleast_2d([1.0,1.0]))

tic = time.perf_counter()
# plot initial data
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
plt.subplots_adjust(left=0.30, bottom=0.30)

# plot just data to get bounds for axes
global d, l, s, p
p = ax.scatter(X.ravel(order='C'), Y.ravel(order='C'), Z_ravel, c='#000000', marker='D', zorder=2)
ax.autoscale(True)
ax.relim()
ax.autoscale_view()
ax.autoscale(False)
p.remove()

# plot again
d = ax.plot_surface(X_pred, Y_pred, Z_pred, color='b', zorder=1, alpha=0.7)
p = ax.scatter(X.ravel(order='C'), Y.ravel(order='C'), Z_ravel, c='#000000', marker='D', zorder=2)
p.set_color(colors)
l = ax.plot_surface(X_pred, Y_pred, computed_answer, color='#00FF00', zorder=2, alpha=0.3)
s = ax.scatter([center_about_extra_coord[0],], [center_about_extra_coord[1],], [computed_answer.ravel(order='C')[center_about_extra_idx],], c='#FF0000', marker='o', zorder=4, edgecolor='black', s=180, alpha=1.0)
ax.set(xlabel='x', ylabel='GMLS approximation')
ax.grid()
ax.margins(x=0)
axcolor = 'lightgoldenrodyellow'

# axes for sliders and radio buttons
ax_location_check = plt.axes([0.05, 0.093, 0.15, 0.15], facecolor='white', frameon=False)
ax_location_x = plt.axes([0.25, 0.175, 0.65, 0.03], facecolor='white')
ax_location_y = plt.axes([0.25, 0.125, 0.65, 0.03], facecolor='white')
ax_weighting_power = plt.axes([0.7, 0.025, 0.2, 0.03], facecolor=axcolor)
ax_num_data_points = plt.axes([0.25, 0.025, 0.2, 0.03], facecolor=axcolor)
ax_epsilon = plt.axes([0.7, 0.075, 0.2, 0.03])
ax_polynomial_order = plt.axes([0.25, 0.075, 0.2, 0.03])
ax_weighting_type = plt.axes([0.015, 0.25, 0.25, 0.15], facecolor=axcolor)
ax_func_type = plt.axes([0.015, 0.40, 0.25, 0.15], facecolor=axcolor)

# sliders
sl_location_check = CheckButtons(ax_location_check, ["",], [True,])
delta_f = 4.0/200
sl_location_x = Slider(ax_location_x, 'Location (x)', valmin=0.0, valmax=4.0, valinit=0.0, valstep=delta_f, color=None, initcolor='black')
sl_location_y = Slider(ax_location_y, 'Location (y)', valmin=0.0, valmax=4.0, valinit=0.0, valstep=delta_f, color=None, initcolor='black')
sl_location_x.set_val(1.0)
sl_location_y.set_val(1.0)
sl_weighting_power = Slider(ax_weighting_power, 'Weighting P.', valmin=0, valmax=5, valinit=3, valstep=1)
sl_num_data_points = Slider(ax_num_data_points, 'Number of Points', valmin=6, valmax=50, valinit=10, valstep=1)
sl_epsilon = Slider(ax_epsilon, 'Epsilon Multipler', valmin=1.0001, valmax=5.0, valinit=1.6, valstep=.1)
sl_polynomial_order = Slider(ax_polynomial_order, 'Polynomial Order', valmin=0, valmax=6, valinit=2, valstep=1)

#radios
rad_weighting_type = RadioButtons(ax_weighting_type, ('Power', 'Cubic Spl.', 'Gaussian'), active=0)
rad_func_type = RadioButtons(ax_func_type, ('x^2*y^2', 'sin(2x)cos(2y)'), active=0)
toc = time.perf_counter()
PRINT_PLOT_TIME and print("Setup graphics in %0.6f seconds"%(toc-tic,))

def update(val):
    global weighting_type
    Z_pred, computed_answer, center_about_extra_idx, center_about_extra_coord, colors = approximate(sl_polynomial_order.val, sl_weighting_power.val, weighting_type, sl_epsilon.val, np.atleast_2d([sl_location_x.val, sl_location_y.val]))
    global d, l, s
    try:
        d.remove()
    except:
        pass
    d = ax.plot_surface(X_pred, Y_pred, Z_pred, color='b', zorder=1,alpha=0.7)
    try:
        l.remove()
    except:
        pass
    l = ax.plot_surface(X_pred, Y_pred, computed_answer, color='#00FF00', zorder=2, alpha=0.3, visible=sl_location_check.get_status()[0])
    try:
        s.remove()
    except:
        pass
    s = ax.scatter([center_about_extra_coord[0],], [center_about_extra_coord[1],], [computed_answer.ravel(order='C')[center_about_extra_idx],], c='#FF0000', marker='o', zorder=4, edgecolor='black', s=180, alpha=1.0)
    if sl_location_check.get_status()[0]:
        p.set_color(colors)
    fig.canvas.draw_idle()
# register objects using update
sl_location_x.on_changed(update)
sl_location_y.on_changed(update)
sl_weighting_power.on_changed(update)
sl_epsilon.on_changed(update)
sl_polynomial_order.on_changed(update)

# called from changefunc
def updatepoints(val):
    global num_data_points, x, y, X, Y, XY_ravel
    num_data_points = sl_num_data_points.val
    x = np.resize(x, new_shape=(num_data_points,))
    y = np.resize(y, new_shape=(num_data_points,))
    x[:] = np.linspace(0,4,num_data_points)
    y[:] = np.linspace(0,4,num_data_points)
    X, Y = np.meshgrid(x,y)
    XY_ravel = np.vstack((X.ravel(order = 'C'),Y.ravel(order = 'C'))).T

def weighting_type_update(label):
    weighting_type_dict = {'Power': 'power', 'Cubic Spl.': 'cubicspline', 'Gaussian' : 'gaussian'}
    global weighting_type
    weighting_type = weighting_type_dict[label]
    if (weighting_type=='power'):
        ax_weighting_power.set_visible(True)
    else:
        ax_weighting_power.set_visible(False)
    update(0)
# register objects using weighting_type_update
rad_weighting_type.on_clicked(weighting_type_update)

def changefunc(label):
    updatepoints(0)
    global weighting_type
    global function
    if type(label)==str:
        func_type_dict = {'x^2*y^2': lambda x,y: pow(x,2)*pow(y,2), 'sin(2x)cos(2y)': lambda x,y: np.sin(2*x)*np.cos(2*y)}
        function = func_type_dict[label]
    global X, Y, Z, Z_ravel
    Z = function(X,Y)
    Z_ravel = Z.ravel(order='C')
    update(0)
    global p, l, s, d
    d.remove()
    l.remove()
    s.remove()
    p.remove()
    p = ax.scatter(X.ravel(order='C'), Y.ravel(order='C'), Z_ravel, c='#000000', marker='D', zorder=2)
    ax.autoscale(True)
    if type(label)==str:
        ax.relim()
        ax.autoscale_view()
    ax.autoscale(False)
    update(0)
    fig.canvas.draw_idle()

# register objects using changefunc
rad_func_type.on_clicked(changefunc)
sl_num_data_points.on_changed(changefunc)

def changelocationviz(label):
    l.set_visible(sl_location_check.get_status()[0])
    fig.canvas.draw_idle()
    if sl_location_check.get_status()[0]:
        update(0)
    else:
        p.set_color('black')
sl_location_check.on_clicked(changelocationviz)

plt.show()
del kp



import pycompadre
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button, RadioButtons, CheckButtons

''' Interactive example of weighting kernels in pycompadre

The weighting kernels used in the GMLS class in the function Wab allow users
to vary two parameters. This tool gives users the ability to vary these two
parameters and weighting type and see the effect visually.
'''

# initialize Kokkos
kp = pycompadre.KokkosParser()

# initialize parameters
wt = pycompadre.WeightingFunctionType.Power
x = np.linspace(-2.0, 2.0, 700)
h = 1.0
p = 2
n = 1

def approximate(wt, h, p, n):
    y = np.array([pycompadre.Wab(xin,h,wt,p,n) for xin in x], dtype='f8')
    return y

# get initial data
y = approximate(wt, h, p, n)

# plot initial data
fig, ax = plt.subplots()
plt.subplots_adjust(left=0.25, bottom=0.25)

# resize to make font more readable when large
fig.set_size_inches(1600/fig.dpi,900/fig.dpi)
global plot_size
plot_size = fig.get_size_inches()*fig.dpi

d, = plt.plot(x, y, c='#0000FF', linewidth=4, zorder=1)
ax.set(xlabel='r', ylabel='',
       title='Kernel Evaluation')
ax.grid()
ax.margins(x=0)
axcolor = 'lightgoldenrodyellow'
ax.autoscale(True)
ax.relim()
ax.autoscale_view()
ax.autoscale(False)

# axes for sliders and radio buttons
ax_p = plt.axes([0.25, 0.125, 0.65, 0.03], facecolor='white')
ax_n = plt.axes([0.25, 0.075, 0.65, 0.03], facecolor='white')
ax_weighting_type = plt.axes([0.015, 0.25, 0.15, 0.35], facecolor=axcolor)

# sliders
delta_f = 4.0/200
sl_p = Slider(ax_p, 'P', valmin=0.0, valmax=6.0, valinit=2, valstep=1, color=None, initcolor='black')
sl_n = Slider(ax_n, 'N', valmin=0.0, valmax=6.0, valinit=1, valstep=1, color=None, initcolor='black')

fontsize=10.0/800*plot_size[0]
t_power = plt.text(0.02, 0.05, '$$\\left(1-\\left(\\frac{|r|}{h}\\right)^N\\right)^P$$', fontsize=fontsize, transform=plt.gcf().transFigure,usetex=True)
t_gaussian = plt.text(0.02, 0.05, '$$\\frac{1}{\\frac{h}{P} \\sqrt(2 \\pi)} e^{\\frac{-1}{2}r^2/{\\left(\\frac{h}{P}\\right)^2}}$$', fontsize=fontsize, transform=plt.gcf().transFigure,usetex=True)
t_cubic = plt.text(0.02, 0.05, '$$((1-\\frac{r}{h})+\\frac{r}{h} (1-\\frac{r}{h}) (1-2 \\frac{r}{h}))$$', fontsize=fontsize, transform=plt.gcf().transFigure,usetex=True)
t_cosine = plt.text(0.02, 0.05, '$$\\cos(\\frac{\\pi r}{2h})$$', fontsize=fontsize, transform=plt.gcf().transFigure,usetex=True)
t_sigmoid = plt.text(0.02, 0.05, '$$\\frac{1}{e^{Pr} + e^{-Pr} + N}$$', fontsize=fontsize, transform=plt.gcf().transFigure,usetex=True)

t_labels = {'Power':t_power, 'Gaussian':t_gaussian, 'Cubic Spl.':t_cubic, 'Cosine':t_cosine, 'Sigmoid':t_sigmoid}
for item in t_labels.keys():
    if item!='Power':
        t_labels[item].update({'visible':False})

#radios
rad_weighting_type = RadioButtons(ax_weighting_type, ('Power', 'Cubic Spl.', 'Cosine', 'Gaussian', 'Sigmoid'), active=0)

def update(val):
    global wt
    y = approximate(wt, h, int(sl_p.val), int(sl_n.val))
    d.set_ydata(y)
    fig.canvas.draw_idle()

# register objects using update
sl_p.on_changed(update)
sl_n.on_changed(update)

def weighting_type_update(label):
    weighting_type_dict = {'Power': pycompadre.WeightingFunctionType.Power, 'Cubic Spl.': pycompadre.WeightingFunctionType.CubicSpline, 'Cosine': pycompadre.WeightingFunctionType.Cosine, 'Gaussian': pycompadre.WeightingFunctionType.Gaussian, 'Sigmoid': pycompadre.WeightingFunctionType.Sigmoid}
    global wt
    wt = weighting_type_dict[label]
    for item in t_labels.keys():
        t_labels[item].update({'visible':False})
    t_labels[label].update({'visible':True})
    if (label=='Power'):
        ax_p.set_visible(True)
        ax_n.set_visible(True)
    elif (label=='Cubic Spl.'):
        ax_p.set_visible(False)
        ax_n.set_visible(False)
    elif (label=='Cosine'):
        ax_p.set_visible(False)
        ax_n.set_visible(False)
    elif (label=='Gaussian'):
        ax_p.set_visible(True)
        ax_n.set_visible(False)
    elif (label=='Sigmoid'):
        ax_p.set_visible(True)
        ax_n.set_visible(True)

    y0 = approximate(wt, h, 0, 0)
    y1 = approximate(wt, h, 0, 6)
    y2 = approximate(wt, h, 6, 0)
    y3 = approximate(wt, h, 6, 6)

    dy0, = ax.plot(x, y0, c='#0000FF', linewidth=4, zorder=1)
    dy1, = ax.plot(x, y1, c='#0000FF', linewidth=4, zorder=1)
    dy2, = ax.plot(x, y2, c='#0000FF', linewidth=4, zorder=1)
    dy3, = ax.plot(x, y3, c='#0000FF', linewidth=4, zorder=1)
    ax.autoscale(True)
    d.set_ydata(y0)
    ax.relim()
    ax.autoscale_view()
    ax.autoscale(False)

    dy0.remove()
    dy1.remove()
    dy2.remove()
    dy3.remove()

    update(0)

# register objects using weighting_type_update
rad_weighting_type.on_clicked(weighting_type_update)

def on_move(event):
    if event.inaxes:
        new_plot_size = fig.get_size_inches()*fig.dpi
        global plot_size
        if (not np.array_equal(new_plot_size, plot_size)):
            for item in t_labels.keys():
                t_labels[item].update({'fontsize':10.0/1600*new_plot_size[0]})
            plot_size = new_plot_size
            fig.canvas.draw_idle()

plt.connect('motion_notify_event', on_move)

plt.show()
del kp
